import json
from unittest.mock import MagicMock, patch

import pytest

from mx_bluesky.common.external_interaction.ispyb.data_model import (
    DataCollectionInfo,
    DataCollectionPositionInfo,
    ScanDataInfo,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_store import (
    IspybIds,
    StoreInIspyb,
)

from ......conftest import (
    EXPECTED_END_TIME,
    EXPECTED_START_TIME,
    TEST_SAMPLE_ID,
)
from ......expeye_helpers import (
    DC_COMMENT_RE,
    DC_RE,
    DCG_RE,
    DCGS_RE,
    DCS_RE,
    POSITION_RE,
    TEST_DATA_COLLECTION_GROUP_ID,
    TEST_DATA_COLLECTION_IDS,
)
from ...xray_centre.test_ispyb_handler import DCG_ID
from .test_gridscan_ispyb_store_3d import TEST_PROPOSAL_REF, TEST_VISIT_NUMBER

EXPECTED_UPDATE_DATA_COLLECTION = {
    "detectorId": 78,
    "axisStart": 0.0,
    "axisRange": 0.1,
    "axisEnd": 180,
    "beamSizeAtSampleX": 1,
    "beamSizeAtSampleY": 1,
    "transmission": 100.0,
    "dataCollectionNumber": 0,
    "detectorDistance": 100.0,
    "exposureTime": 0.1,
    "imageDirectory": "/tmp/",
    "imagePrefix": "file_name",
    "imageSuffix": "h5",
    "numberOfPasses": 1,
    "overlap": 0,
    "omegaStart": 0,
    "startImageNumber": 1,
    "wavelength": 123.98419840550369,
    "xBeam": 150.0,
    "yBeam": 160.0,
    "xtalSnapshotFullPath1": "test_1_y",
    "xtalSnapshotFullPath2": "test_2_y",
    "xtalSnapshotFullPath3": "test_3_y",
    "startTime": EXPECTED_START_TIME,
    "fileTemplate": "file_name_1_master.h5",
    "numberOfImages": 1800,
    "kappaStart": 0,
}

EXPECTED_BEGIN_DATA_COLLECTION = EXPECTED_UPDATE_DATA_COLLECTION | {
    "comments": "Hyperion rotation scan",
}


@pytest.fixture
@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
def scan_data_info_for_begin():
    return ScanDataInfo(
        data_collection_info=DataCollectionInfo(
            omega_start=0.0,
            data_collection_number=0,
            xtal_snapshot1="test_1_y",
            xtal_snapshot2="test_2_y",
            xtal_snapshot3="test_3_y",
            n_images=1800,
            axis_range=0.1,
            axis_end=180.0,
            kappa_start=0.0,
            parent_id=None,
            visit_string="cm31105-4",
            ispyb_detector_id=78,
            axis_start=0.0,
            beamsize_at_samplex=1.0,
            beamsize_at_sampley=1.0,
            transmission=100.0,
            comments="Hyperion rotation scan",
            detector_distance=100.0,
            exp_time=0.1,
            imgdir="/tmp/",
            file_template="file_name_1_master.h5",
            imgprefix="file_name",
            imgsuffix="h5",
            n_passes=1,
            overlap=0,
            start_image_number=1,
            wavelength=123.98419840550369,
            xbeam=150.0,
            ybeam=160.0,
            synchrotron_mode=None,
            undulator_gap1=None,
            start_time="2024-02-08 14:03:59",
        ),
        data_collection_id=None,
        data_collection_position_info=None,
        data_collection_grid_info=None,
    )


@pytest.fixture
def scan_data_info_for_update(scan_data_info_for_begin):
    return ScanDataInfo(
        data_collection_info=DataCollectionInfo(
            omega_start=0.0,
            data_collection_number=0,
            xtal_snapshot1="test_1_y",
            xtal_snapshot2="test_2_y",
            xtal_snapshot3="test_3_y",
            n_images=1800,
            axis_range=0.1,
            axis_end=180.0,
            kappa_start=0.0,
            parent_id=None,
            visit_string="cm31105-4",
            ispyb_detector_id=78,
            axis_start=0.0,
            slitgap_vertical=1.0,
            slitgap_horizontal=1.0,
            beamsize_at_samplex=1.0,
            beamsize_at_sampley=1.0,
            transmission=100.0,
            detector_distance=100.0,
            exp_time=0.1,
            imgdir="/tmp/",
            file_template="file_name_1_master.h5",
            imgprefix="file_name",
            imgsuffix="h5",
            n_passes=1,
            overlap=0,
            flux=10.0,
            start_image_number=1,
            wavelength=123.98419840550369,
            xbeam=150.0,
            ybeam=160.0,
            synchrotron_mode="test",
            undulator_gap1=None,
            start_time="2024-02-08 14:03:59",
        ),
        data_collection_position_info=DataCollectionPositionInfo(
            pos_x=10.0, pos_y=20.0, pos_z=30.0
        ),
        data_collection_grid_info=None,
    )


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
def test_begin_deposition(
    mock_ispyb_conn,
    dummy_ispyb,
    dummy_rotation_data_collection_group_info,
    scan_data_info_for_begin,
):
    assert scan_data_info_for_begin.data_collection_info.parent_id is None

    assert dummy_ispyb.begin_deposition(
        dummy_rotation_data_collection_group_info, [scan_data_info_for_begin]
    ) == IspybIds(
        data_collection_ids=(TEST_DATA_COLLECTION_IDS[0],),
        data_collection_group_id=TEST_DATA_COLLECTION_GROUP_ID,
    )

    assert (
        scan_data_info_for_begin.data_collection_info.parent_id
        == TEST_DATA_COLLECTION_GROUP_ID
    )

    create_dcg_request = mock_ispyb_conn.calls_for(DCGS_RE)[0].request
    assert mock_ispyb_conn.match(create_dcg_request, DCGS_RE, 2) == TEST_PROPOSAL_REF
    assert (
        int(mock_ispyb_conn.match(create_dcg_request, DCGS_RE, 3)) == TEST_VISIT_NUMBER
    )
    assert json.loads(create_dcg_request.body) == {
        "experimentType": "SAD",
        "sampleId": TEST_SAMPLE_ID,
    }
    create_dc_request = mock_ispyb_conn.calls_for(DCS_RE)[0].request
    assert (
        int(mock_ispyb_conn.match(create_dc_request, DCS_RE, 2))
        == TEST_DATA_COLLECTION_GROUP_ID
    )
    assert json.loads(create_dc_request.body) == EXPECTED_BEGIN_DATA_COLLECTION


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
def test_begin_deposition_with_group_id_updates_but_doesnt_insert(
    mock_ispyb_conn,
    dummy_ispyb,
    dummy_rotation_data_collection_group_info,
    scan_data_info_for_begin,
):
    scan_data_info_for_begin.data_collection_info.parent_id = (
        TEST_DATA_COLLECTION_GROUP_ID
    )

    assert dummy_ispyb.begin_deposition(
        dummy_rotation_data_collection_group_info, [scan_data_info_for_begin]
    ) == IspybIds(
        data_collection_ids=(TEST_DATA_COLLECTION_IDS[0],),
        data_collection_group_id=TEST_DATA_COLLECTION_GROUP_ID,
    )
    assert len(mock_ispyb_conn.calls_for(DCGS_RE)) == 0
    update_dcg_request = mock_ispyb_conn.calls_for(DCG_RE)[0].request
    assert (
        int(mock_ispyb_conn.match(update_dcg_request, DCG_RE, 2))
        == TEST_DATA_COLLECTION_GROUP_ID
    )
    assert json.loads(update_dcg_request.body) == {
        "experimentType": "SAD",
        "sampleId": TEST_SAMPLE_ID,
    }
    create_dc_request = mock_ispyb_conn.calls_for(DCS_RE)[0].request
    assert (
        int(mock_ispyb_conn.match(create_dc_request, DCS_RE, 2))
        == TEST_DATA_COLLECTION_GROUP_ID
    )
    assert json.loads(create_dc_request.body) == EXPECTED_BEGIN_DATA_COLLECTION


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
def test_begin_deposition_with_alternate_experiment_type(
    mock_ispyb_conn,
    dummy_ispyb,
    dummy_rotation_data_collection_group_info,
    scan_data_info_for_begin,
):
    dummy_rotation_data_collection_group_info.experiment_type = "Characterization"
    assert dummy_ispyb.begin_deposition(
        dummy_rotation_data_collection_group_info,
        [scan_data_info_for_begin],
    ) == IspybIds(
        data_collection_ids=(TEST_DATA_COLLECTION_IDS[0],),
        data_collection_group_id=TEST_DATA_COLLECTION_GROUP_ID,
    )
    create_dcg_request = mock_ispyb_conn.calls_for(DCGS_RE)[0].request
    assert mock_ispyb_conn.match(create_dcg_request, DCGS_RE, 2) == TEST_PROPOSAL_REF
    assert (
        int(mock_ispyb_conn.match(create_dcg_request, DCGS_RE, 3)) == TEST_VISIT_NUMBER
    )
    assert json.loads(create_dcg_request.body) == {
        "experimentType": "Characterization",
        "sampleId": TEST_SAMPLE_ID,
    }


@patch(
    "mx_bluesky.common.external_interaction.ispyb.ispyb_store.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
def test_update_deposition(
    mock_ispyb_conn,
    dummy_ispyb,
    dummy_rotation_data_collection_group_info,
    scan_data_info_for_begin,
    scan_data_info_for_update,
):
    ispyb_ids = dummy_ispyb.begin_deposition(
        dummy_rotation_data_collection_group_info, [scan_data_info_for_begin]
    )
    assert len(mock_ispyb_conn.calls_for(DCGS_RE)) == 1
    assert len(mock_ispyb_conn.calls_for(DCS_RE)) == 1
    assert len(mock_ispyb_conn.calls_for(DC_RE)) == 0

    scan_data_info_for_update.data_collection_info.parent_id = (
        ispyb_ids.data_collection_group_id
    )
    scan_data_info_for_update.data_collection_id = ispyb_ids.data_collection_ids[0]

    assert dummy_ispyb.update_deposition(
        ispyb_ids,
        [scan_data_info_for_update],
    ) == IspybIds(
        data_collection_group_id=TEST_DATA_COLLECTION_GROUP_ID,
        data_collection_ids=(TEST_DATA_COLLECTION_IDS[0],),
    )
    assert len(mock_ispyb_conn.calls_for(DCGS_RE)) == 1
    update_dc_request = mock_ispyb_conn.dc_calls_for(DC_RE)[0]
    assert update_dc_request.dcid == TEST_DATA_COLLECTION_IDS[0]
    assert update_dc_request.body == EXPECTED_UPDATE_DATA_COLLECTION | {
        "synchrotronMode": "test",
        "slitGapVertical": 1,
        "slitGapHorizontal": 1,
        "flux": 10,
    }

    create_pos_request = mock_ispyb_conn.dc_calls_for(POSITION_RE)[0]
    assert create_pos_request.dcid == TEST_DATA_COLLECTION_IDS[0]
    assert create_pos_request.body == {
        "posX": 10,
        "posY": 20,
        "posZ": 30,
    }


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
def test_update_deposition_with_group_id_updates(
    mock_ispyb_conn,
    dummy_rotation_data_collection_group_info,
    scan_data_info_for_begin,
    scan_data_info_for_update,
    dummy_ispyb: StoreInIspyb,
):
    scan_data_info_for_begin.data_collection_info.parent_id = (
        TEST_DATA_COLLECTION_GROUP_ID
    )
    ispyb_ids = dummy_ispyb.begin_deposition(
        dummy_rotation_data_collection_group_info, [scan_data_info_for_begin]
    )
    assert len(mock_ispyb_conn.calls_for(DCGS_RE)) == 0
    assert len(mock_ispyb_conn.calls_for(DCS_RE)) == 1
    assert len(mock_ispyb_conn.calls_for(DC_RE)) == 0

    scan_data_info_for_update.data_collection_info.parent_id = (
        ispyb_ids.data_collection_group_id
    )
    scan_data_info_for_update.data_collection_id = ispyb_ids.data_collection_ids[0]

    assert dummy_ispyb.update_deposition(
        ispyb_ids,
        [scan_data_info_for_update],
    ) == IspybIds(
        data_collection_group_id=TEST_DATA_COLLECTION_GROUP_ID,
        data_collection_ids=(TEST_DATA_COLLECTION_IDS[0],),
    )
    assert len(mock_ispyb_conn.calls_for(DCGS_RE)) == 0
    assert len(mock_ispyb_conn.calls_for(DCG_ID)) == 0
    update_dc_request = mock_ispyb_conn.dc_calls_for(DC_RE)[0]
    assert update_dc_request.dcid == TEST_DATA_COLLECTION_IDS[0]
    assert update_dc_request.body == EXPECTED_UPDATE_DATA_COLLECTION | {
        "synchrotronMode": "test",
        "slitGapVertical": 1,
        "slitGapHorizontal": 1,
        "flux": 10,
    }

    update_position_request = mock_ispyb_conn.dc_calls_for(POSITION_RE)[0]
    assert update_position_request.dcid == TEST_DATA_COLLECTION_IDS[0]
    assert update_position_request.body == {
        "posX": 10,
        "posY": 20,
        "posZ": 30,
    }


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
@patch(
    "mx_bluesky.common.external_interaction.ispyb.ispyb_store.get_current_time_string",
)
def test_end_deposition_happy_path(
    get_current_time,
    mock_ispyb_conn,
    dummy_ispyb,
    dummy_rotation_data_collection_group_info,
    scan_data_info_for_begin,
    scan_data_info_for_update,
):
    ispyb_ids = dummy_ispyb.begin_deposition(
        dummy_rotation_data_collection_group_info, [scan_data_info_for_begin]
    )
    scan_data_info_for_update.data_collection_info.parent_id = (
        ispyb_ids.data_collection_group_id
    )
    scan_data_info_for_update.data_collection_id = ispyb_ids.data_collection_ids[0]
    ispyb_ids = dummy_ispyb.update_deposition(
        ispyb_ids,
        [scan_data_info_for_update],
    )
    assert len(mock_ispyb_conn.calls_for(DC_RE)) == 1

    get_current_time.return_value = EXPECTED_END_TIME
    dummy_ispyb.end_deposition(ispyb_ids, "success", "Test succeeded")
    dcids_to_append_comment_reqs = {
        rq.dcid: rq for rq in mock_ispyb_conn.dc_calls_for(DC_COMMENT_RE)
    }
    assert dcids_to_append_comment_reqs[TEST_DATA_COLLECTION_IDS[0]].body == {
        "comments": " DataCollection Successful reason: Test succeeded",
    }
    assert len(mock_ispyb_conn.calls_for(DC_RE)) == 2
    update_dc_request = mock_ispyb_conn.dc_calls_for(DC_RE)[1]
    assert update_dc_request.dcid == TEST_DATA_COLLECTION_IDS[0]
    assert update_dc_request.body == {
        "endTime": EXPECTED_END_TIME,
        "runStatus": "DataCollection Successful",
    }


def test_store_rotation_scan_failures(mock_ispyb_conn, dummy_ispyb: StoreInIspyb):
    ispyb_ids = IspybIds(
        data_collection_group_id=TEST_DATA_COLLECTION_GROUP_ID,
    )
    with pytest.raises(AssertionError):
        dummy_ispyb.end_deposition(ispyb_ids, "", "")


@pytest.mark.parametrize(
    "mock_ispyb_conn, dcgid",
    [[{"dcg_id": dcg_id}, dcg_id] for dcg_id in [2, 45, 61, 88, 13, 25]],
    indirect=["mock_ispyb_conn"],
)
def test_store_rotation_scan_uses_supplied_dcgid(
    mock_ispyb_conn,
    dcgid,
    dummy_ispyb,
    dummy_rotation_data_collection_group_info,
    scan_data_info_for_begin,
    scan_data_info_for_update,
):
    scan_data_info_for_begin.data_collection_info.parent_id = dcgid
    ispyb_ids = dummy_ispyb.begin_deposition(
        dummy_rotation_data_collection_group_info, [scan_data_info_for_begin]
    )
    assert ispyb_ids.data_collection_group_id == dcgid
    update_dcg_request = mock_ispyb_conn.calls_for(DCG_RE)[0].request
    assert int(mock_ispyb_conn.match(update_dcg_request, DCG_RE, 2)) == dcgid
    assert json.loads(update_dcg_request.body) == {
        "experimentType": "SAD",
        "sampleId": TEST_SAMPLE_ID,
    }
    scan_data_info_for_update.data_collection_id = ispyb_ids.data_collection_ids[0]
    assert (
        dummy_ispyb.update_deposition(
            ispyb_ids,
            [scan_data_info_for_update],
        ).data_collection_group_id
        == dcgid
    )



from __future__ import annotations

import json
from pathlib import Path

import bluesky.preprocessors as bpp
import pytest
from bluesky.run_engine import RunEngine
from numpy.testing import assert_allclose
from ophyd_async.core import SignalR
from ophyd_async.epics.motor import Motor

from mx_bluesky.beamlines.i24.jungfrau_commissioning.callbacks.metadata_writer import (
    READING_DUMP_FILENAME,
    JsonMetadataWriter,
)
from mx_bluesky.beamlines.i24.jungfrau_commissioning.experiment_plans.rotation_scan_plan import (
    RotationScanComposite,
)
from mx_bluesky.beamlines.i24.parameters.constants import PlanNameConstants
from mx_bluesky.common.experiment_plans.inner_plans.read_hardware import (
    read_hardware_plan,
)
from mx_bluesky.common.parameters.rotation import SingleRotationScan
from tests.unit_tests.beamlines.i24.jungfrau_commissioning.utils import (
    get_good_single_rotation_params,
)


async def test_metadata_writer_produces_correct_output(
    run_engine: RunEngine, tmp_path, rotation_composite: RotationScanComposite
):
    params = get_good_single_rotation_params(tmp_path)
    metadata_writer = JsonMetadataWriter()

    wavelength = 1
    energy = 1
    det_z = 3

    await rotation_composite.dcm.wavelength_in_a.set(wavelength)
    await rotation_composite.dcm.energy_in_keV.set(energy)
    await rotation_composite.det_stage.z.set(det_z)
    await rotation_composite.jungfrau._writer.file_path.set(tmp_path)

    expected_output = {
        "wavelength_in_a": wavelength,
        "energy_kev": energy,
        "detector_distance_mm": det_z,
        "angular_increment_deg": 0.1,
    }
    run_engine(
        _do_metadata_writing_read(
            [
                rotation_composite.dcm.energy_in_keV,
                rotation_composite.dcm.wavelength_in_a,
                rotation_composite.det_stage.z,
                rotation_composite.jungfrau._writer.file_path,
            ],
            params,
            metadata_writer,
        )
    )

    assert metadata_writer.final_path == tmp_path

    with open(Path(tmp_path) / READING_DUMP_FILENAME) as f:
        actual_output = json.load(f)
    assert expected_output.keys() == actual_output.keys()
    for key in actual_output:
        assert_allclose(actual_output[key], expected_output[key])


async def test_assertion_error_if_no_jf_path_found(
    run_engine: RunEngine, tmp_path, rotation_composite: RotationScanComposite
):
    params = get_good_single_rotation_params(tmp_path)
    metadata_writer = JsonMetadataWriter()
    wavelength = 1
    energy = 1
    det_z = 3

    await rotation_composite.dcm.wavelength_in_a.set(wavelength)
    await rotation_composite.dcm.energy_in_keV.set(energy)
    await rotation_composite.det_stage.z.set(det_z)

    with pytest.raises(AssertionError, match="No detector writer path was found"):
        run_engine(
            _do_metadata_writing_read(
                [
                    rotation_composite.dcm.energy_in_keV,
                    rotation_composite.dcm.wavelength_in_a,
                    rotation_composite.det_stage.z,
                    rotation_composite.jungfrau._writer.file_path,
                ],
                params,
                metadata_writer,
            )
        )


def _do_metadata_writing_read(
    signals: list[SignalR | Motor],
    params: SingleRotationScan,
    writer: JsonMetadataWriter,
):
    @bpp.subs_decorator([writer])
    @bpp.set_run_key_decorator(PlanNameConstants.ROTATION_MAIN)
    @bpp.run_decorator(
        md={
            "subplan_name": PlanNameConstants.ROTATION_MAIN,
            "rotation_scan_params": params.model_dump_json(),
        }
    )
    def _inner_read():
        yield from read_hardware_plan(
            signals,
            PlanNameConstants.ROTATION_DEVICE_READ,
        )

    yield from _inner_read()



from unittest.mock import patch

import pytest

from mx_bluesky.beamlines.i24.web_gui_plans.oav_plans import (
    Direction,
    FocusDirection,
    MoveSize,
    focus_on_oav_view,
    move_block_on_arrow_click,
    move_nudge_on_arrow_click,
    move_window_on_arrow_click,
)


@pytest.mark.parametrize(
    "direction, expected_value",
    [
        ("up", 3.1750),
        ("left", -3.1750),
        ("right", 3.1750),
        ("down", -3.1750),
    ],
)
def test_move_block_on_arrow_click(direction, expected_value, pmac, run_engine):
    with patch(
        "mx_bluesky.beamlines.i24.web_gui_plans.oav_plans.bps.abs_set",
    ) as mock_abs_set:
        run_engine(move_block_on_arrow_click(Direction(direction), pmac))
        if direction in ["left", "right"]:
            mock_abs_set.assert_any_call(pmac.x, expected_value, wait=True)
        else:
            mock_abs_set.assert_any_call(pmac.y, expected_value, wait=True)


@pytest.mark.parametrize(
    "direction, move_size, expected_value",
    [
        ("up", "small", 0.1250),
        ("up", "big", 0.3750),
        ("left", "small", -0.1250),
        ("left", "big", -0.3750),
        ("right", "small", 0.1250),
        ("right", "big", 0.3750),
        ("down", "small", -0.1250),
        ("down", "big", -0.3750),
    ],
)
def test_move_window_on_arrow_click(
    direction, move_size, expected_value, pmac, run_engine
):
    with patch(
        "mx_bluesky.beamlines.i24.web_gui_plans.oav_plans.bps.abs_set",
    ) as mock_abs_set:
        run_engine(
            move_window_on_arrow_click(Direction(direction), MoveSize(move_size), pmac)
        )
        if direction in ["left", "right"]:
            mock_abs_set.assert_any_call(pmac.x, expected_value, wait=True)
        else:
            mock_abs_set.assert_any_call(pmac.y, expected_value, wait=True)


@pytest.mark.parametrize(
    "direction, move_size, expected_value",
    [
        ("up", "small", 0.0010),
        ("up", "big", 0.0060),
        ("left", "small", -0.0010),
        ("left", "big", -0.0060),
        ("right", "small", 0.0010),
        ("right", "big", 0.0060),
        ("down", "small", -0.0010),
        ("down", "big", -0.0060),
    ],
)
def test_move_nudge_on_arrow_click(
    direction, move_size, expected_value, pmac, run_engine
):
    with patch(
        "mx_bluesky.beamlines.i24.web_gui_plans.oav_plans.bps.abs_set",
    ) as mock_abs_set:
        run_engine(
            move_nudge_on_arrow_click(Direction(direction), MoveSize(move_size), pmac)
        )
        if direction in ["left", "right"]:
            mock_abs_set.assert_any_call(pmac.x, expected_value, wait=True)
        else:
            mock_abs_set.assert_any_call(pmac.y, expected_value, wait=True)


@pytest.mark.parametrize(
    "direction, move_size, expected_value",
    [
        ("in", "small", -0.0200),
        ("in", "big", -0.1200),
        ("out", "small", 0.0200),
        ("out", "big", 0.1200),
    ],
)
async def test_focus_on_oav_view(
    direction, move_size, expected_value, pmac, run_engine
):
    run_engine(focus_on_oav_view(FocusDirection(direction), MoveSize(move_size), pmac))
    assert await pmac.z.user_readback.get_value() == expected_value



import json
from unittest.mock import MagicMock, patch

from mx_bluesky.hyperion.external_interaction.callbacks.rotation.ispyb_callback import (
    RotationISPyBCallback,
)

from ......conftest import (
    EXPECTED_END_TIME,
    EXPECTED_START_TIME,
    replace_all_tmp_paths,
)
from ......expeye_helpers import (
    DC_COMMENT_RE,
    DC_RE,
    DCG_RE,
    DCGS_RE,
    DCS_RE,
    POSITION_RE,
    TEST_DATA_COLLECTION_GROUP_ID,
    TEST_DATA_COLLECTION_IDS,
)
from .....common.external_interaction.callbacks.ispyb.test_gridscan_ispyb_store_3d import (
    TEST_PROPOSAL_REF,
    TEST_VISIT_NUMBER,
)

TEST_SAMPLE_ID = 123456

EXPECTED_DATA_COLLECTION = {
    "axisStart": 0.0,
    "axisRange": 0.1,
    "axisEnd": -180,
    "comments": "test",
    "chiStart": 23.85,
    "dataCollectionNumber": 1,
    "detectorDistance": 100.0,
    "exposureTime": 0.1,
    "imageDirectory": "{tmp_data}/123456/",
    "imagePrefix": "file_name",
    "imageSuffix": "h5",
    "numberOfPasses": 1,
    "overlap": 0,
    "omegaStart": 0,
    "startImageNumber": 1,
    "xBeam": 150.0,
    "yBeam": 160.0,
    "startTime": EXPECTED_START_TIME,
    "fileTemplate": "file_name_1_master.h5",
    "numberOfImages": 1800,
    "sampleId": 123456,
}


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
def test_activity_gated_start(
    mock_ispyb_conn, test_rotation_start_outer_document, tmp_path
):
    callback = RotationISPyBCallback()

    callback.activity_gated_start(test_rotation_start_outer_document)
    create_dcg_request = mock_ispyb_conn.calls_for(DCGS_RE)[0].request
    assert mock_ispyb_conn.match(create_dcg_request, DCGS_RE, 2) == TEST_PROPOSAL_REF
    assert (
        int(mock_ispyb_conn.match(create_dcg_request, DCGS_RE, 3)) == TEST_VISIT_NUMBER
    )
    assert json.loads(create_dcg_request.body) == {
        "experimentType": "SAD",
        "sampleId": TEST_SAMPLE_ID,
    }
    create_dc_request = mock_ispyb_conn.calls_for(DCS_RE)[0].request
    assert (
        int(mock_ispyb_conn.match(create_dc_request, DCS_RE, 2))
        == TEST_DATA_COLLECTION_GROUP_ID
    )
    assert json.loads(create_dc_request.body) == replace_all_tmp_paths(
        EXPECTED_DATA_COLLECTION, tmp_path
    )


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
def test_hardware_read_events(
    mock_ispyb_conn, test_rotation_start_outer_document, test_event_data
):
    callback = RotationISPyBCallback()
    callback.activity_gated_start(test_rotation_start_outer_document)  # pyright: ignore
    callback.activity_gated_start(
        test_event_data.test_rotation_start_main_document  # pyright: ignore
    )

    callback.activity_gated_descriptor(
        test_event_data.test_descriptor_document_pre_data_collection
    )
    callback.activity_gated_event(
        test_event_data.test_event_document_pre_data_collection
    )
    assert len(mock_ispyb_conn.calls_for(DCG_RE)) == 0
    update_dc_request = mock_ispyb_conn.dc_calls_for(DC_RE)[0]
    assert update_dc_request.dcid == TEST_DATA_COLLECTION_IDS[0]
    assert update_dc_request.body == {
        "slitGapHorizontal": 0.1234,
        "slitGapVertical": 0.2345,
        "synchrotronMode": "User",
        "undulatorGap1": 1.234,
        "resolution": 1.1830593331191241,
        "wavelength": 1.11647184541378,
    }
    append_comments_req = list(mock_ispyb_conn.dc_calls_for(DC_COMMENT_RE))[0]
    assert append_comments_req.dcid == TEST_DATA_COLLECTION_IDS[0]
    assert append_comments_req.body == {
        "comments": " Sample position (µm): (158, 24, 3)"
    }

    expected_data = test_event_data.test_event_document_pre_data_collection["data"]
    create_position_request = mock_ispyb_conn.dc_calls_for(POSITION_RE)[0]
    assert create_position_request.dcid == TEST_DATA_COLLECTION_IDS[0]
    assert create_position_request.body == {
        "posX": expected_data["gonio-x"],
        "posY": expected_data["gonio-y"],
        "posZ": expected_data["gonio-z"],
    }


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
def test_flux_read_events(
    mock_ispyb_conn, test_rotation_start_outer_document, test_event_data
):
    callback = RotationISPyBCallback()
    callback.activity_gated_start(test_rotation_start_outer_document)  # pyright: ignore
    callback.activity_gated_start(
        test_event_data.test_rotation_start_main_document  # pyright: ignore
    )
    callback.activity_gated_descriptor(
        test_event_data.test_descriptor_document_pre_data_collection
    )
    callback.activity_gated_event(
        test_event_data.test_event_document_pre_data_collection
    )
    assert len(mock_ispyb_conn.calls_for(DC_RE)) == 1
    callback.activity_gated_descriptor(
        test_event_data.test_descriptor_document_during_data_collection
    )
    callback.activity_gated_event(
        test_event_data.test_rotation_event_document_during_data_collection
    )

    assert len(mock_ispyb_conn.calls_for(DCG_RE)) == 0
    update_dc_request = mock_ispyb_conn.dc_calls_for(DC_RE)[1]
    assert update_dc_request.dcid == TEST_DATA_COLLECTION_IDS[0]
    assert update_dc_request.body == {
        "beamSizeAtSampleX": 0.05,
        "beamSizeAtSampleY": 0.02,
        "wavelength": 1.11647184541378,
        "transmission": 98,
        "flux": 9.81,
        "resolution": 1.1830593331191241,
        "detectorMode": "FULL",
        "detectorId": 78,
    }


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
def test_oav_rotation_snapshot_triggered_event(
    mock_ispyb_conn, test_rotation_start_outer_document, test_event_data
):
    callback = RotationISPyBCallback()
    callback.activity_gated_start(test_rotation_start_outer_document)  # pyright: ignore
    callback.activity_gated_start(
        test_event_data.test_rotation_start_main_document  # pyright: ignore
    )
    callback.activity_gated_descriptor(
        test_event_data.test_descriptor_document_oav_rotation_snapshot
    )

    assert len(mock_ispyb_conn.calls_for(DC_RE)) == 0
    i = 0
    for snapshot in [
        {"filename": "snapshot_0", "colname": "xtalSnapshotFullPath1"},
        {"filename": "snapshot_90", "colname": "xtalSnapshotFullPath2"},
        {"filename": "snapshot_180", "colname": "xtalSnapshotFullPath3"},
        {"filename": "snapshot_270", "colname": "xtalSnapshotFullPath4"},
    ]:
        event_doc = dict(test_event_data.test_event_document_oav_rotation_snapshot)
        event_doc["data"]["oav-snapshot-last_saved_path"] = snapshot["filename"]  # type: ignore
        callback.activity_gated_event(event_doc)  # type: ignore
        update_dc_request = mock_ispyb_conn.dc_calls_for(DC_RE)[i]
        assert update_dc_request.dcid == TEST_DATA_COLLECTION_IDS[0]
        assert update_dc_request.body == {
            snapshot["colname"]: snapshot["filename"],
        }
        i += 1

    assert len(mock_ispyb_conn.calls_for(DCG_RE)) == 0


@patch(
    "mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping.get_current_time_string",
    new=MagicMock(return_value=EXPECTED_START_TIME),
)
def test_activity_gated_stop(
    mock_ispyb_conn,
    test_rotation_start_outer_document,
    test_rotation_stop_outer_document,
    test_event_data,
):
    callback = RotationISPyBCallback()
    callback.activity_gated_start(test_rotation_start_outer_document)  # pyright: ignore
    callback.activity_gated_start(
        test_event_data.test_rotation_start_main_document  # pyright: ignore
    )

    assert len(mock_ispyb_conn.calls_for(DCG_RE)) == 0
    assert len(mock_ispyb_conn.calls_for(DC_RE)) == 0

    with patch(
        "mx_bluesky.common.external_interaction.ispyb.ispyb_store.get_current_time_string",
        new=MagicMock(return_value=EXPECTED_END_TIME),
    ):
        callback.activity_gated_stop(test_event_data.test_rotation_stop_main_document)
        callback.activity_gated_stop(test_rotation_stop_outer_document)

    update_comment_req = mock_ispyb_conn.dc_calls_for(DC_COMMENT_RE)[0]
    assert update_comment_req.dcid == TEST_DATA_COLLECTION_IDS[0]
    assert update_comment_req.body == {
        "comments": " DataCollection Successful reason: Test succeeded"
    }
    update_dc_request = mock_ispyb_conn.dc_calls_for(DC_RE)[0]
    assert update_dc_request.dcid == TEST_DATA_COLLECTION_IDS[0]
    assert update_dc_request.body == {
        "endTime": EXPECTED_END_TIME,
        "runStatus": "DataCollection Successful",
    }
    assert len(mock_ispyb_conn.dc_calls_for(DC_RE)) == 1


def test_comment_correct_after_hardware_read(
    mock_ispyb_conn, test_rotation_start_outer_document, test_event_data
):
    callback = RotationISPyBCallback()
    test_rotation_start_outer_document["mx_bluesky_parameters"] = (
        test_rotation_start_outer_document["mx_bluesky_parameters"].replace(
            '"comment":"test"', '"comment":"a lovely unit test"'
        )
    )
    callback.activity_gated_start(test_rotation_start_outer_document)  # pyright: ignore
    callback.activity_gated_start(
        test_event_data.test_rotation_start_main_document  # pyright: ignore
    )

    create_dc_request = mock_ispyb_conn.dc_calls_for(DCS_RE)[0]
    assert create_dc_request.body["comments"] == "a lovely unit test"

    callback.activity_gated_descriptor(
        test_event_data.test_descriptor_document_pre_data_collection
    )
    callback.activity_gated_event(
        test_event_data.test_event_document_pre_data_collection
    )
    update_dc_request = mock_ispyb_conn.dc_calls_for(DC_RE)[0]
    assert update_dc_request.dcid == TEST_DATA_COLLECTION_IDS[0]
    assert update_dc_request.body == {
        "slitGapHorizontal": 0.1234,
        "slitGapVertical": 0.2345,
        "synchrotronMode": "User",
        "undulatorGap1": 1.234,
        "resolution": 1.1830593331191241,
        "wavelength": 1.11647184541378,
    }
    update_dc_comment_req = mock_ispyb_conn.dc_calls_for(DC_COMMENT_RE)[0]
    assert update_dc_comment_req.dcid == TEST_DATA_COLLECTION_IDS[0]
    assert update_dc_comment_req.body == {
        "comments": " Sample position (µm): (158, 24, 3)"
    }



from __future__ import annotations

import dataclasses

from dodal.devices.zebra.zebra import RotationDirection
from dodal.utils import get_beamline_name

from mx_bluesky.common.parameters.constants import RotationParamConstants
from mx_bluesky.common.parameters.rotation import SingleRotationScan
from mx_bluesky.common.utils.log import LOGGER

DEFAULT_DIRECTION = RotationDirection.NEGATIVE
DEFAULT_MAX_VELOCITY = 120
# Use a slightly larger time to acceleration than EPICS as it's better to be cautious
ACCELERATION_MARGIN = 1.5


@dataclasses.dataclass
class RotationMotionProfile:
    start_scan_deg: float
    start_motion_deg: float
    scan_width_deg: float
    shutter_time_s: float
    direction: RotationDirection
    speed_for_rotation_deg_s: float
    acceleration_offset_deg: float
    shutter_opening_deg: float
    total_exposure_s: float
    distance_to_move_deg: float
    max_velocity_deg_s: float


def calculate_motion_profile(
    params: SingleRotationScan,
    motor_time_to_speed_s: float,
    max_velocity_deg_s: float,
) -> RotationMotionProfile:
    """Calculates the various numbers needed for motions in the rotation scan.
    Rotates through "scan width" plus twice an "offset" to take into account
    acceleration at the start and deceleration at the end, plus the number of extra
    degrees of rotation needed to make sure the fast shutter has fully opened before the
    detector trigger is sent.
    See https://github.com/DiamondLightSource/hyperion/wiki/rotation-scan-geometry
    for a simple pictorial explanation."""

    assert params.rotation_increment_deg > 0

    direction = params.rotation_direction
    start_scan_deg = params.omega_start_deg

    if RotationParamConstants.OMEGA_FLIP:
        # If omega_flip is True then the motor omega axis is inverted with respect to the
        # coordinate system.
        start_scan_deg = -start_scan_deg
        direction = (
            direction.POSITIVE
            if direction == direction.NEGATIVE
            else direction.NEGATIVE
        )

    num_images = params.num_images
    shutter_time_s = params.shutter_opening_time_s
    image_width_deg = params.rotation_increment_deg
    exposure_time_s = params.exposure_time_s
    motor_time_to_speed_s *= ACCELERATION_MARGIN

    LOGGER.info("Calculating rotation scan motion profile:")
    LOGGER.info(
        f"{num_images=}, {shutter_time_s=}, {image_width_deg=}, {exposure_time_s=}, {direction=}"
    )

    scan_width_deg = num_images * params.rotation_increment_deg
    LOGGER.info(f"{scan_width_deg=} = {num_images=} * {params.rotation_increment_deg=}")

    speed_for_rotation_deg_s = image_width_deg / exposure_time_s
    LOGGER.info("speed_for_rotation_deg_s = image_width_deg / exposure_time_s")
    LOGGER.info(
        f"{speed_for_rotation_deg_s=} = {image_width_deg=} / {exposure_time_s=}"
    )

    acceleration_offset_deg = motor_time_to_speed_s * speed_for_rotation_deg_s
    LOGGER.info(
        f"{acceleration_offset_deg=} = {motor_time_to_speed_s=} * {speed_for_rotation_deg_s=}"
    )

    start_motion_deg = start_scan_deg - (acceleration_offset_deg * direction.multiplier)
    LOGGER.info(
        f"{start_motion_deg=} = {start_scan_deg=} - ({acceleration_offset_deg=} * {direction.multiplier=})"
    )

    shutter_opening_deg = speed_for_rotation_deg_s * shutter_time_s
    LOGGER.info(
        f"{shutter_opening_deg=} = {speed_for_rotation_deg_s=} * {shutter_time_s=}"
    )

    shutter_opening_deg = speed_for_rotation_deg_s * shutter_time_s
    LOGGER.info(
        f"{shutter_opening_deg=} = {speed_for_rotation_deg_s=} * {shutter_time_s=}"
    )

    total_exposure_s = num_images * exposure_time_s
    LOGGER.info(f"{total_exposure_s=} = {num_images=} * {exposure_time_s=}")

    distance_to_move_deg = (
        scan_width_deg + shutter_opening_deg + acceleration_offset_deg * 2
    ) * direction.multiplier
    LOGGER.info(
        f"{distance_to_move_deg=} = ({scan_width_deg=} + {shutter_opening_deg=} + {acceleration_offset_deg=} * 2) * {direction=})"
    )

    # See https://github.com/DiamondLightSource/mx-bluesky/issues/1224
    if get_beamline_name() == "i24":
        acceleration_offset_deg = 10

    return RotationMotionProfile(
        start_scan_deg=start_scan_deg,
        start_motion_deg=start_motion_deg,
        scan_width_deg=scan_width_deg,
        shutter_time_s=shutter_time_s,
        direction=direction,
        speed_for_rotation_deg_s=speed_for_rotation_deg_s,
        acceleration_offset_deg=acceleration_offset_deg,
        shutter_opening_deg=shutter_opening_deg,
        total_exposure_s=total_exposure_s,
        distance_to_move_deg=distance_to_move_deg,
        max_velocity_deg_s=max_velocity_deg_s,
    )



from collections.abc import Sequence
from typing import Any, Literal

from mx_bluesky.beamlines.i24.serial.fixed_target.ft_utils import ChipType
from mx_bluesky.beamlines.i24.serial.parameters.experiment_parameters import (
    ChipDescription,
)
from mx_bluesky.beamlines.i24.serial.setup_beamline import caget, pv

OXFORD_BLOCKS_PVS = [f"BL24I-MO-IOC-13:GP{i}" for i in range(11, 75)]


class EmptyMapError(Exception):
    pass


def get_chip_format(
    chip_type: ChipType,
    format: Sequence[int | float] | None = None,
    origin: Literal["edm", "web"] = "edm",
) -> ChipDescription:
    """Get the default parameter values for the requested chip type.

    For an Oxford-type chip, the default values are hard coded as the dimensions are
    always the same. For a Custom chip instead, the number of steps and step size in
    each direction must be entered through the GUI - web or edm. If the collection is
    run through the edm, the values will be read from the general purpose PVs set on
    there. If instead the plan is run from the web UI, the values will be passed in the
    form of a list/tuple of 4 values.

    Args:
        chip_type (ChipType): Chip in use
        custom_format (Sequence[int | float], optional): Number and size of steps input
            from the web ui. Format should be: [int, int, float, float].
            Defaults to None.
        origin (str, optional): UI in use, can be either web or edm. Defaults to edm.
    """
    defaults: dict[str, int | float] = {}
    match chip_type:
        case ChipType.Oxford:
            defaults["x_num_steps"] = defaults["y_num_steps"] = 20
            defaults["x_step_size"] = defaults["y_step_size"] = 0.125
            defaults["x_blocks"] = defaults["y_blocks"] = 8
            defaults["b2b_horz"] = defaults["b2b_vert"] = 0.800
        case ChipType.OxfordInner:
            defaults["x_num_steps"] = defaults["y_num_steps"] = 25
            defaults["x_step_size"] = defaults["y_step_size"] = 0.600
            defaults["x_blocks"] = defaults["y_blocks"] = 1
            defaults["b2b_horz"] = defaults["b2b_vert"] = 0.0
        case ChipType.Minichip:
            defaults["x_num_steps"] = defaults["y_num_steps"] = 20
            defaults["x_step_size"] = defaults["y_step_size"] = 0.125
            defaults["x_blocks"] = defaults["y_blocks"] = 1
            defaults["b2b_horz"] = defaults["b2b_vert"] = 0.0
        case ChipType.Custom:
            if origin == "edm":
                defaults["x_num_steps"] = int(caget(pv.ioc13_gp6))
                defaults["y_num_steps"] = int(caget(pv.ioc13_gp7))
                defaults["x_step_size"] = float(caget(pv.ioc13_gp8))
                defaults["y_step_size"] = float(caget(pv.ioc13_gp99))
            else:
                # NOTE Test for WEB GUI
                if not format:
                    raise ValueError("Format for custom chip not passed")
                defaults["x_num_steps"] = format[0]
                defaults["y_num_steps"] = format[1]
                defaults["x_step_size"] = format[2]
                defaults["y_step_size"] = format[3]
            defaults["x_blocks"] = defaults["y_blocks"] = 1
            defaults["b2b_horz"] = defaults["b2b_vert"] = 0.0
        case ChipType.MISP:
            defaults["x_num_steps"] = defaults["y_num_steps"] = 78
            defaults["x_step_size"] = defaults["y_step_size"] = 0.1193
            defaults["x_blocks"] = defaults["y_blocks"] = 1
            defaults["b2b_horz"] = defaults["b2b_vert"] = 0.0
    chip_params: dict[str, Any] = {"chip_type": chip_type, **defaults}
    return ChipDescription(**chip_params)


def get_chip_map() -> list[int]:
    """Return a list of blocks (the 'chip map') to be collected on an Oxford type chip \
        when using lite mapping."""
    chipmap = []
    for n, block_pv in enumerate(OXFORD_BLOCKS_PVS):
        block_val = int(caget(block_pv))
        if block_val == 1:
            chipmap.append(n + 1)
    if len(chipmap) == 0:
        raise EmptyMapError("No blocks selected for Lite map.")
    return chipmap



import argparse
from enum import StrEnum
from pathlib import Path

from pydantic.dataclasses import dataclass

from mx_bluesky._version import version
from mx_bluesky.hyperion.parameters.constants import HyperionConstants


class HyperionMode(StrEnum):
    UDC = "udc"
    SUPERVISOR = "supervisor"


@dataclass
class HyperionArgs:
    mode: HyperionMode
    dev_mode: bool = False
    client_config: str | None = None
    supervisor_config: str | None = None


@dataclass
class CallbackArgs:
    dev_mode: bool = False
    watchdog_port: int = HyperionConstants.HYPERION_PORT
    stomp_config: Path | None = None


def _add_callback_relevant_args(parser: argparse.ArgumentParser) -> None:
    """adds arguments relevant to hyperion-callbacks."""
    parser.add_argument(
        "--dev",
        action="store_true",
        help="Use dev options, such as local graylog instances",
    )


def parse_callback_args() -> CallbackArgs:
    """Parse the CLI arguments for the watchdog port and dev mode into a CallbackArgs instance."""
    parser = argparse.ArgumentParser()
    _add_callback_relevant_args(parser)
    parser.add_argument(
        "--watchdog-port",
        type=int,
        help="Liveness port for callbacks to ping regularly",
    )
    parser.add_argument(
        "--stomp-config",
        type=Path,
        default=None,
        help="Specify config yaml for the STOMP backend (default is 0MQ)",
    )
    args = parser.parse_args()
    return CallbackArgs(
        dev_mode=args.dev,
        watchdog_port=args.watchdog_port,
        stomp_config=args.stomp_config,
    )


def parse_cli_args() -> HyperionArgs:
    """Parses all arguments relevant to hyperion.
    Returns:
         an HyperionArgs dataclass with the fields: (dev_mode: bool)"""
    parser = argparse.ArgumentParser()
    _add_callback_relevant_args(parser)
    parser.add_argument(
        "--version",
        help="Print hyperion version string",
        action="version",
        version=version,
    )
    parser.add_argument(
        "--mode",
        help="Launch in the specified mode (default is 'udc')",
        default=HyperionMode.UDC,
        type=HyperionMode,
        choices=HyperionMode.__members__.values(),
    )
    parser.add_argument(
        "--client-config", help="Specify the blueapi client configuration file."
    )
    parser.add_argument(
        "--supervisor-config",
        help="Specify the supervisor bluesky context configuration file.",
    )
    args = parser.parse_args()
    return HyperionArgs(
        dev_mode=args.dev or False,
        mode=args.mode,
        supervisor_config=args.supervisor_config,
        client_config=args.client_config,
    )



from pathlib import Path
from unittest.mock import MagicMock, call, patch

import numpy as np
import pytest
from bluesky.run_engine import RunEngine
from bluesky.simulators import assert_message_and_return_remaining
from bluesky.utils import Msg
from dodal.devices.aperturescatterguard import (
    ApertureValue,
)
from dodal.devices.zocalo import ZocaloResults
from dodal.devices.zocalo.zocalo_results import _NO_SAMPLE_ID
from ophyd_async.core import completed_status, set_mock_value
from ophyd_async.fastcs.panda import DatasetTable, PandaHdf5DatasetType

from mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan import (
    BeamlineSpecificFGSFeatures,
    common_flyscan_xray_centre,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback import (
    GridscanISPyBCallback,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.nexus_callback import (
    GridscanNexusFileCallback,
)
from mx_bluesky.common.parameters.constants import (
    DeviceSettingsConstants,
)
from mx_bluesky.common.parameters.device_composites import FlyScanEssentialDevices
from mx_bluesky.hyperion.experiment_plans.hyperion_flyscan_xray_centre_plan import (
    SmargonSpeedError,
)
from mx_bluesky.hyperion.parameters.device_composites import (
    HyperionFlyScanXRayCentreComposite,
)
from mx_bluesky.hyperion.parameters.gridscan import (
    HyperionSpecifiedThreeDGridScan,
)
from tests.conftest import (
    RunEngineSimulator,
)

from ....conftest import TEST_RESULT_LARGE, TestData, simulate_xrc_result
from ...conftest import (
    mock_zocalo_trigger,
    modified_store_grid_scan_mock,
)

ReWithSubs = tuple[RunEngine, tuple[GridscanNexusFileCallback, GridscanISPyBCallback]]


class CompleteError(Exception):
    # To avoid having to run through the entire plan during tests
    pass


def _custom_msg(command_name: str):
    return lambda *args, **kwargs: iter([Msg(command_name)])


@pytest.fixture
def fgs_composite_with_panda_pcap(
    hyperion_flyscan_xrc_composite: HyperionFlyScanXRayCentreComposite,
):
    capture_table = DatasetTable(name=["name"], dtype=[PandaHdf5DatasetType.FLOAT_64])
    set_mock_value(hyperion_flyscan_xrc_composite.panda.data.datasets, capture_table)

    return hyperion_flyscan_xrc_composite


@patch(
    "mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback.StoreInIspyb",
    modified_store_grid_scan_mock,
)
class TestFlyscanXrayCentrePlan:
    @patch(
        "dodal.devices.aperturescatterguard.ApertureScatterguard._safe_move_within_datacollection_range",
        side_effect=lambda: completed_status(),
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.run_gridscan",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.change_aperture_then_move_plan.move_x_y_z",
        autospec=True,
    )
    @pytest.mark.skip(
        reason="TODO mx-bluesky 231 aperture size should be determined from absolute size not box size"
    )
    def test_results_adjusted_and_passed_to_move_xyz(
        self,
        move_x_y_z: MagicMock,
        move_aperture: MagicMock,
        run_gridscan: MagicMock,
        hyperion_flyscan_xrc_composite: HyperionFlyScanXRayCentreComposite,
        hyperion_fgs_params: HyperionSpecifiedThreeDGridScan,
        run_engine_with_subs: ReWithSubs,
        beamline_specific: BeamlineSpecificFGSFeatures,
    ):
        run_engine, _ = run_engine_with_subs

        for result in [
            TestData.test_result_large,
            TestData.test_result_medium,
            TestData.test_result_small,
        ]:
            mock_zocalo_trigger(hyperion_flyscan_xrc_composite.zocalo, result)
            run_engine(
                common_flyscan_xray_centre(
                    hyperion_flyscan_xrc_composite,
                    hyperion_fgs_params,
                    beamline_specific,
                )
            )

        aperture_scatterguard = hyperion_flyscan_xrc_composite.aperture_scatterguard
        large = aperture_scatterguard._loaded_positions[ApertureValue.LARGE]
        medium = aperture_scatterguard._loaded_positions[ApertureValue.MEDIUM]
        ap_call_large = call(large, ApertureValue.LARGE)
        ap_call_medium = call(medium, ApertureValue.MEDIUM)

        move_aperture.assert_has_calls([ap_call_large, ap_call_large, ap_call_medium])

        mv_to_centre = call(
            hyperion_flyscan_xrc_composite.gonio,
            0.05,
            pytest.approx(0.15),
            0.25,
            wait=True,
        )
        move_x_y_z.assert_has_calls(
            [mv_to_centre, mv_to_centre, mv_to_centre], any_order=True
        )

    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.run_gridscan",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.change_aperture_then_move_plan.move_x_y_z",
        autospec=True,
    )
    async def test_when_gridscan_finished_then_dev_shm_disabled(
        self,
        move_xyz: MagicMock,
        run_gridscan: MagicMock,
        sim_run_engine: RunEngineSimulator,
        hyperion_fgs_params: HyperionSpecifiedThreeDGridScan,
        hyperion_flyscan_xrc_composite: FlyScanEssentialDevices,
        beamline_specific: BeamlineSpecificFGSFeatures,
        zocalo: ZocaloResults,
    ):
        hyperion_flyscan_xrc_composite.eiger.odin.fan.dev_shm_enable.sim_put(1)  # type: ignore
        sim_run_engine.add_read_handler_for(
            zocalo.centre_of_mass, [np.array([6.0, 6.0, 6.0])]
        )
        sim_run_engine.add_read_handler_for(zocalo.max_voxel, [np.array([5, 5, 5])])
        sim_run_engine.add_read_handler_for(zocalo.max_count, [123456])
        sim_run_engine.add_read_handler_for(zocalo.n_voxels, [321])
        sim_run_engine.add_read_handler_for(zocalo.total_count, [999999])
        sim_run_engine.add_read_handler_for(
            zocalo.bounding_box, [np.array([[3, 3, 3], [9, 9, 9]])]
        )
        sim_run_engine.add_read_handler_for(zocalo.sample_id, [_NO_SAMPLE_ID])
        msgs = sim_run_engine.simulate_plan(
            common_flyscan_xray_centre(
                hyperion_flyscan_xrc_composite,
                hyperion_fgs_params,
                beamline_specific,
            )
        )

        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "set"
            and msg.obj is hyperion_flyscan_xrc_composite.eiger.odin.fan.dev_shm_enable
            and msg.args[0] == 0,
        )
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "wait"
            and msg.kwargs["group"] == msgs[0].kwargs["group"],
        )

    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.kickoff_and_complete_gridscan",
    )
    def test_if_smargon_speed_over_limit_then_log_error(
        self,
        mock_kickoff_and_complete: MagicMock,
        fgs_params_use_panda: HyperionSpecifiedThreeDGridScan,
        hyperion_flyscan_xrc_composite: FlyScanEssentialDevices,
        beamline_specific: BeamlineSpecificFGSFeatures,
        run_engine: RunEngine,
    ):
        fgs_params_use_panda.x_step_size_um = 10000
        fgs_params_use_panda.detector_params.exposure_time_s = 0.01

        # this exception should only be raised if we're using the panda
        with pytest.raises(SmargonSpeedError):
            run_engine(
                common_flyscan_xray_centre(
                    hyperion_flyscan_xrc_composite,
                    fgs_params_use_panda,
                    beamline_specific,
                )
            )

    @patch(
        "mx_bluesky.hyperion.device_setup_plans.setup_panda.arm_panda_for_gridscan",
        new=MagicMock(side_effect=_custom_msg("arm_panda")),
    )
    @patch(
        "mx_bluesky.hyperion.experiment_plans.hyperion_flyscan_xray_centre_plan.disarm_panda_for_gridscan",
        new=MagicMock(side_effect=_custom_msg("disarm_panda")),
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.run_gridscan",
        new=MagicMock(side_effect=_custom_msg("do_gridscan")),
    )
    @patch(
        "mx_bluesky.hyperion.experiment_plans.hyperion_flyscan_xray_centre_plan.set_panda_directory",
        side_effect=_custom_msg("set_panda_directory"),
    )
    @patch("mx_bluesky.hyperion.device_setup_plans.setup_panda.load_panda_from_yaml")
    def test_flyscan_xray_centre_sets_directory_stages_arms_disarms_unstages_the_panda(
        self,
        mock_load_panda: MagicMock,
        mock_set_panda_directory: MagicMock,
        fgs_params_use_panda: HyperionSpecifiedThreeDGridScan,
        fgs_composite_with_panda_pcap: HyperionFlyScanXRayCentreComposite,
        sim_run_engine: RunEngineSimulator,
        beamline_specific: BeamlineSpecificFGSFeatures,
        tmp_path: Path,
    ):
        sim_run_engine.add_read_handler_for(
            fgs_composite_with_panda_pcap.gonio.x.max_velocity, 10
        )
        simulate_xrc_result(
            sim_run_engine, fgs_composite_with_panda_pcap.zocalo, TEST_RESULT_LARGE
        )

        msgs = sim_run_engine.simulate_plan(
            common_flyscan_xray_centre(
                fgs_composite_with_panda_pcap, fgs_params_use_panda, beamline_specific
            )
        )

        mock_set_panda_directory.assert_called_with(tmp_path / "xraycentring/123456")
        mock_load_panda.assert_called_once_with(
            DeviceSettingsConstants.PANDA_FLYSCAN_SETTINGS_DIR,
            DeviceSettingsConstants.PANDA_FLYSCAN_SETTINGS_FILENAME,
            fgs_composite_with_panda_pcap.panda,
        )

        msgs = assert_message_and_return_remaining(
            msgs, lambda msg: msg.command == "set_panda_directory"
        )
        msgs = assert_message_and_return_remaining(
            msgs, lambda msg: msg.command == "stage" and msg.obj.name == "panda"
        )
        msgs = assert_message_and_return_remaining(
            msgs, lambda msg: msg.command == "arm_panda"
        )
        msgs = assert_message_and_return_remaining(
            msgs, lambda msg: msg.command == "do_gridscan"
        )
        msgs = assert_message_and_return_remaining(
            msgs, lambda msg: msg.command == "disarm_panda"
        )
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "unstage"
            and msg.obj.name == "panda"
            and msg.kwargs["group"] == "panda_flyscan_tidy",
        )
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "wait"
            and msg.kwargs["group"] == "panda_flyscan_tidy",
        )



from functools import partial

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
from bluesky.preprocessors import run_decorator
from bluesky.utils import MsgGenerator
from dodal.devices.beamlines.i24.aperture import AperturePositions
from dodal.devices.beamlines.i24.beamstop import BeamstopPositions
from dodal.devices.beamlines.i24.commissioning_jungfrau import CommissioningJungfrau
from dodal.devices.beamlines.i24.dual_backlight import BacklightPositions
from dodal.devices.hutch_shutter import ShutterState
from dodal.devices.zebra.zebra import ArmDemand, I24Axes, Zebra
from dodal.devices.zebra.zebra_controlled_shutter import ZebraShutter
from ophyd_async.fastcs.jungfrau import (
    GainMode,
    create_jungfrau_external_triggering_info,
)
from pydantic import BaseModel, field_validator

from mx_bluesky.beamlines.i24.jungfrau_commissioning.callbacks.metadata_writer import (
    JsonMetadataWriter,
)
from mx_bluesky.beamlines.i24.jungfrau_commissioning.composites import (
    RotationScanComposite,
)
from mx_bluesky.beamlines.i24.jungfrau_commissioning.plan_stubs.plan_utils import (
    JF_COMPLETE_GROUP,
    fly_jungfrau,
)
from mx_bluesky.beamlines.i24.parameters.constants import (
    PlanNameConstants,
)
from mx_bluesky.common.device_setup_plans.setup_zebra_and_shutter import (
    setup_zebra_for_rotation,
    tidy_up_zebra_after_rotation_scan,
)
from mx_bluesky.common.experiment_plans.inner_plans.read_hardware import (
    read_hardware_plan,
)
from mx_bluesky.common.experiment_plans.rotation.rotation_utils import (
    RotationMotionProfile,
    calculate_motion_profile,
)
from mx_bluesky.common.parameters.components import get_param_version
from mx_bluesky.common.parameters.constants import (
    USE_NUMTRACKER,
    PlanGroupCheckpointConstants,
)
from mx_bluesky.common.parameters.rotation import (
    SingleRotationScan,
)
from mx_bluesky.common.utils.log import LOGGER

READING_DUMP_FILENAME = "collection_info.json"

# Should be read from config file, see
# https://github.com/DiamondLightSource/mx-bluesky/issues/1502
JF_DET_STAGE_Y_POSITION_MM = 730
DEFAULT_DETECTOR_DISTANCE_MM = 200


class ExternalRotationScanParams(BaseModel):
    transmission_fractions: list[float]
    exposure_time_s: float
    omega_start_deg: float = 0
    rotation_increment_per_image_deg: float = 0.1
    filename: str = "rotations"
    detector_distance_mm: float = DEFAULT_DETECTOR_DISTANCE_MM
    sample_id: int

    @field_validator("transmission_fractions")
    @classmethod
    def validate_transmission_fractions(cls, values):
        for v in values:
            if not 0 <= v <= 1:
                raise ValueError(
                    f"All transmission fractions must be between 0 and 1; got {v}"
                )
        return values


def _get_internal_rotation_params(
    entry_params: ExternalRotationScanParams, transmission: float
) -> SingleRotationScan:
    return SingleRotationScan(
        sample_id=entry_params.sample_id,
        visit=USE_NUMTRACKER,  # See https://github.com/DiamondLightSource/mx-bluesky/issues/1527
        parameter_model_version=get_param_version(),
        file_name=entry_params.filename,
        transmission_frac=transmission,
        exposure_time_s=entry_params.exposure_time_s,
        storage_directory=USE_NUMTRACKER,
    )


class HutchClosedError(Exception): ...


def rotation_scan_plan(
    composite: RotationScanComposite, params: ExternalRotationScanParams
) -> MsgGenerator:
    """BlueAPI entry point for i24 JF rotation scans"""

    for transmission in params.transmission_fractions:
        rotation_params = _get_internal_rotation_params(params, transmission)
        yield from single_rotation_plan(composite, rotation_params)


def set_up_beamline_for_rotation(
    composite: RotationScanComposite,
    det_z_mm: float,
    transmission_frac: float,
):
    """Check hutch is open, then, in parallel, move backlight in,
    move aperture in, move beamstop out and move det stages in. Wait for this parallel
    move to finish."""

    hutch_shutter_state: ShutterState = yield from bps.rd(
        composite.hutch_shutter.status
    )
    LOGGER.info(f"Hutch shutter: {hutch_shutter_state}")
    if hutch_shutter_state != ShutterState.OPEN:
        LOGGER.error(f"Hutch shutter is not open! State is {hutch_shutter_state}")
        raise HutchClosedError(
            f"Hutch shutter is not open! State is {hutch_shutter_state}"
        )

    LOGGER.info(
        "Making sure aperture and beamstop are in, detector stages are in position, backlight is out, and transmission is set..."
    )
    yield from bps.mv(
        composite.aperture.position,
        AperturePositions.IN,
        composite.beamstop.pos_select,
        BeamstopPositions.DATA_COLLECTION,
        composite.det_stage.y,
        JF_DET_STAGE_Y_POSITION_MM,
        composite.backlight.backlight_position,
        BacklightPositions.OUT,
        composite.det_stage.z,
        det_z_mm,
        composite.attenuator,
        transmission_frac,
    )


def single_rotation_plan(
    composite: RotationScanComposite,
    params: SingleRotationScan,
):
    """A stub plan to collect diffraction images from a sample continuously rotating
    about a fixed axis - for now this axis is limited to omega.
    Needs additional setup of the sample environment and a wrapper to clean up."""

    @bpp.set_run_key_decorator(PlanNameConstants.SINGLE_ROTATION_SCAN)
    @run_decorator()
    def _plan_in_run_decorator():
        if not params.detector_distance_mm:
            LOGGER.info(
                f"Using default detector distance of  {DEFAULT_DETECTOR_DISTANCE_MM} mm"
            )
            params.detector_distance_mm = DEFAULT_DETECTOR_DISTANCE_MM

        yield from set_up_beamline_for_rotation(
            composite, params.detector_distance_mm, params.transmission_frac
        )

        # This value isn't actually used, see https://github.com/DiamondLightSource/mx-bluesky/issues/1224
        _motor_time_to_speed = 1
        _max_velocity_deg_s = yield from bps.rd(composite.gonio.omega.max_velocity)

        motion_values = calculate_motion_profile(
            params, _motor_time_to_speed, _max_velocity_deg_s
        )

        # Callback which intercepts read documents and writes to json file,
        # used for saving device metadata
        metadata_writer = JsonMetadataWriter()

        @bpp.subs_decorator([metadata_writer])
        @bpp.set_run_key_decorator(PlanNameConstants.ROTATION_MAIN)
        @bpp.run_decorator(
            md={
                "subplan_name": PlanNameConstants.ROTATION_MAIN,
                "scan_points": [params.scan_points],
                "rotation_scan_params": params.model_dump_json(),
                "detector_file_template": params.file_name,
            }
        )
        def _rotation_scan_plan(
            motion_values: RotationMotionProfile,
            composite: RotationScanComposite,
        ):
            _jf_trigger_info = create_jungfrau_external_triggering_info(
                params.num_images, params.detector_params.exposure_time_s
            )

            axis = composite.gonio.omega

            # can move to start as fast as possible
            yield from bps.abs_set(
                axis.velocity, motion_values.max_velocity_deg_s, wait=True
            )
            LOGGER.info(f"Moving omega to start value, {motion_values.start_scan_deg=}")
            yield from bps.abs_set(
                axis,
                motion_values.start_motion_deg,
                group=PlanGroupCheckpointConstants.ROTATION_READY_FOR_DC,
            )

            yield from setup_zebra_for_rotation(
                composite.zebra,
                composite.sample_shutter,
                axis=I24Axes.OMEGA,
                start_angle=motion_values.start_scan_deg,
                scan_width=motion_values.scan_width_deg,
                direction=motion_values.direction,
                shutter_opening_deg=motion_values.shutter_opening_deg,
                shutter_opening_s=motion_values.shutter_time_s,
            )

            yield from bps.wait(PlanGroupCheckpointConstants.ROTATION_READY_FOR_DC)

            # Get ready for the actual scan
            yield from bps.abs_set(
                axis.velocity, motion_values.speed_for_rotation_deg_s, wait=True
            )
            yield from bps.abs_set(composite.zebra.pc.arm, ArmDemand.ARM, wait=True)

            # Should check topup gate here, but not yet implemented,
            # see https://github.com/DiamondLightSource/mx-bluesky/issues/1501

            # Read hardware after preparing jungfrau so that device metadata output from callback is correct
            # Whilst metadata is being written in bluesky we need to access the private writer here
            read_hardware_partial = partial(
                read_hardware_plan,
                [
                    composite.dcm.energy_in_keV,
                    composite.dcm.wavelength_in_a,
                    composite.det_stage.z,
                    composite.jungfrau._writer.file_path,  # noqa: SLF001 N
                ],
                PlanNameConstants.ROTATION_DEVICE_READ,
            )

            yield from fly_jungfrau(
                composite.jungfrau,
                _jf_trigger_info,
                GainMode.DYNAMIC,
                wait=False,
                log_on_percentage_prefix="Jungfrau rotation scan triggers received",
                read_hardware_after_prepare_plan=read_hardware_partial,
            )

            LOGGER.info("Executing rotation scan")
            yield from bps.rel_set(
                axis,
                motion_values.distance_to_move_deg,
                wait=False,
                group=JF_COMPLETE_GROUP,
            )

            LOGGER.info(
                "Waiting for omega to finish moving and for Jungfrau to receive correct number of triggers"
            )
            yield from bps.wait(group=JF_COMPLETE_GROUP)

        yield from bpp.finalize_wrapper(
            _rotation_scan_plan(motion_values, composite),
            final_plan=partial(
                _cleanup_plan,
                composite.zebra,
                composite.jungfrau,
                composite.sample_shutter,
            ),
        )

    yield from _plan_in_run_decorator()


def _cleanup_plan(
    zebra: Zebra,
    jf: CommissioningJungfrau,
    zebra_shutter: ZebraShutter,
    group="rotation cleanup",
):
    LOGGER.info("Tidying up Zebra and Jungfrau...")
    yield from bps.unstage(jf, group=group)
    yield from tidy_up_zebra_after_rotation_scan(
        zebra, zebra_shutter, group=group, wait=False
    )
    yield from bps.wait(group=group)



import configparser
from dataclasses import dataclass
from enum import StrEnum
from typing import Any, Literal

from event_model.documents import Event
from requests import JSONDecodeError, patch, post
from requests.auth import AuthBase

from mx_bluesky.common.external_interaction.ispyb.data_model import (
    DataCollectionGridInfo,
    DataCollectionGroupInfo,
    DataCollectionInfo,
    DataCollectionPositionInfo,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_utils import (
    get_current_time_string,
    get_ispyb_config,
)
from mx_bluesky.common.utils.exceptions import ISPyBDepositionNotMadeError

RobotActionID = int


class BearerAuth(AuthBase):
    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers["authorization"] = "Bearer " + self.token
        return r


def _get_base_url_and_token() -> tuple[str, str]:
    config = configparser.ConfigParser()
    conf = get_ispyb_config()
    config.read(conf)
    expeye_config = config["expeye"]
    return expeye_config["url"], expeye_config["token"]


def _send_and_get_response(auth, url, data, send_func, query_params=None) -> dict:
    response = send_func(url, auth=auth, json=data, params=query_params)
    if not response.ok:
        try:
            resp_txt = str(response.json())
        except JSONDecodeError:
            resp_txt = str(response)
        raise ISPyBDepositionNotMadeError(
            f"Could not write {data} to {url}: {resp_txt}"
        )
    return response.json()


@dataclass
class BLSample:
    container_id: int
    bl_sample_id: int
    bl_sample_status: str | None


class BLSampleStatus(StrEnum):
    # The sample has been loaded
    LOADED = "LOADED"
    # Problem with the sample e.g. pin too long/short
    ERROR_SAMPLE = "ERROR - sample"
    # Any other general error
    ERROR_BEAMLINE = "ERROR - beamline"


assert all(len(value) <= 20 for value in BLSampleStatus), (
    "Column size limit of 20 for BLSampleStatus"
)


def create_update_data_from_event_doc(
    mapping: dict[str, str], event: Event
) -> dict[str, Any]:
    """Given a mapping between bluesky event data and an event itself this function will
    create a dict that can be used to update exp-eye."""
    event_data = event["data"]
    return {
        target_key: event_data[source_key]
        for source_key, target_key in mapping.items()
        if source_key in event_data
    }


class ExpeyeInteraction:
    """Exposes functionality from the Expeye core API"""

    CREATE_ROBOT_ACTION = "/proposals/{proposal}/sessions/{visit_number}/robot-actions"
    UPDATE_ROBOT_ACTION = "/robot-actions/{action_id}"

    def __init__(self) -> None:
        url, token = _get_base_url_and_token()
        self._base_url = url
        self._auth = BearerAuth(token)

    def start_robot_action(
        self,
        action_type: Literal["LOAD", "UNLOAD"],
        proposal_reference: str,
        visit_number: int,
        sample_id: int,
    ) -> RobotActionID:
        """Create a robot action entry in ispyb.

        Args:
            action_type ("LOAD" | "UNLOAD"): The robot action being performed
            proposal_reference (str): The proposal of the experiment e.g. cm37235
            visit_number (int): The visit number for the proposal, usually this can be
                                found added to the end of the proposal e.g. the data for
                                visit number 2 of proposal cm37235 is in cm37235-2
            sample_id (int): The id of the sample in the database

        Returns:
            RobotActionID: The id of the robot load action that is created
        """
        url = self._base_url + self.CREATE_ROBOT_ACTION.format(
            proposal=proposal_reference, visit_number=visit_number
        )

        data = {
            "startTimestamp": get_current_time_string(),
            "actionType": action_type,
            "sampleId": sample_id,
        }
        response = _send_and_get_response(self._auth, url, data, post)
        return response["robotActionId"]

    def update_robot_action(
        self,
        action_id: RobotActionID,
        data: dict[str, Any],
    ):
        """Update an existing robot action to contain additional info.

        Args:
            action_id (RobotActionID): The id of the action to update
            data (dict): The data to update with, where the keys match those expected
                         by exp-eye.
        """
        url = self._base_url + self.UPDATE_ROBOT_ACTION.format(action_id=action_id)
        _send_and_get_response(self._auth, url, data, patch)

    def end_robot_action(self, action_id: RobotActionID, status: str, reason: str):
        """Finish an existing robot action, providing final information about how it went

        Args:
            action_id (RobotActionID): The action to finish.
            status (str): The status of the action at the end, "success" for success,
                          otherwise error
            reason (str): If the status is in error than the reason for that error
        """
        url = self._base_url + self.UPDATE_ROBOT_ACTION.format(action_id=action_id)

        run_status = "SUCCESS" if status == "success" else "ERROR"

        data = {
            "endTimestamp": get_current_time_string(),
            "status": run_status,
            "message": reason[:255] if reason else "",
        }
        _send_and_get_response(self._auth, url, data, patch)

    def update_sample_status(
        self, bl_sample_id: int, bl_sample_status: BLSampleStatus
    ) -> BLSample:
        """Update the blSampleStatus of a sample.
        Args:
            bl_sample_id: The sample ID
            bl_sample_status: The sample status
            status_message: An optional message
        Returns:
             The updated sample
        """
        data = {"blSampleStatus": (str(bl_sample_status))}
        response = _send_and_get_response(
            self._auth, self._base_url + f"/samples/{bl_sample_id}", data, patch
        )
        return self._sample_from_json(response)

    def _sample_from_json(self, response) -> BLSample:
        return BLSample(
            bl_sample_id=response["blSampleId"],
            bl_sample_status=response["blSampleStatus"],
            container_id=response["containerId"],
        )

    def create_data_group(
        self, proposal_reference: str, visit_number: int, data: DataCollectionGroupInfo
    ) -> int:
        response = _send_and_get_response(
            self._auth,
            self._base_url + f"/proposals/{proposal_reference}/sessions/"
            f"{visit_number}/data-groups",
            _data_collection_group_info_to_json(data),
            post,
        )
        return response["dataCollectionGroupId"]

    def update_data_group(self, group_id: int, data: DataCollectionGroupInfo):
        _send_and_get_response(
            self._auth,
            self._base_url + f"/data-groups/{group_id}",
            _data_collection_group_info_to_json(data),
            patch,
        )

    def create_data_collection(self, group_id: int, data: DataCollectionInfo) -> int:
        response = _send_and_get_response(
            self._auth,
            self._base_url + f"/data-groups/{group_id}/data-collections",
            _data_collection_info_to_json(data),
            post,
        )
        return response["dataCollectionId"]

    def update_data_collection(
        self,
        data_collection_id: int,
        data: DataCollectionInfo,
        append_comment: bool = False,
    ):
        _send_and_get_response(
            self._auth,
            self._base_url + f"/data-collections/{data_collection_id}",
            _data_collection_info_to_json(data),
            patch,
            {"appendComment": "true"} if append_comment else None,
        )

    def create_position(
        self, data_collection_id: int, data: DataCollectionPositionInfo
    ):
        _send_and_get_response(
            self._auth,
            self._base_url + f"/data-collections/{data_collection_id}/position",
            _position_info_to_json(data),
            post,
        )

    def create_grid(self, data_collection_id: int, data: DataCollectionGridInfo) -> int:
        response = _send_and_get_response(
            self._auth,
            self._base_url + f"/data-collections/{data_collection_id}/grids",
            _grid_info_to_json(data),
            post,
        )
        return response["gridInfoId"]


def _none_to_absent(json: dict) -> dict:
    for key in [key for key in json if json[key] is None]:
        del json[key]
    return json


def _data_collection_group_info_to_json(data: DataCollectionGroupInfo) -> dict:
    return _none_to_absent(
        {
            "experimentType": data.experiment_type,
            "sampleId": data.sample_id,
            "actualSampleBarcode": data.sample_barcode,
            "comments": data.comments,
        }
    )


def _data_collection_info_to_json(data: DataCollectionInfo) -> dict:
    return _none_to_absent(
        {
            "omegaStart": data.omega_start,
            "dataCollectionNumber": data.data_collection_number,
            "xtalSnapshotFullPath1": data.xtal_snapshot1,
            "xtalSnapshotFullPath2": data.xtal_snapshot2,
            "xtalSnapshotFullPath3": data.xtal_snapshot3,
            "xtalSnapshotFullPath4": data.xtal_snapshot4,
            "numberOfImages": data.n_images,
            "axisRange": data.axis_range,
            "axisEnd": data.axis_end,
            "chiStart": data.chi_start,
            "kappaStart": data.kappa_start,
            "detectorId": data.ispyb_detector_id,
            "axisStart": data.axis_start,
            "slitGapVertical": data.slitgap_vertical,
            "slitGapHorizontal": data.slitgap_horizontal,
            "beamSizeAtSampleX": data.beamsize_at_samplex,
            "beamSizeAtSampleY": data.beamsize_at_sampley,
            "transmission": data.transmission,
            "comments": data.comments,
            "detectorDistance": data.detector_distance,
            "exposureTime": data.exp_time,
            "imageDirectory": data.imgdir,
            "fileTemplate": data.file_template,
            "imagePrefix": data.imgprefix,
            "imageSuffix": data.imgsuffix,
            "numberOfPasses": data.n_passes,
            "overlap": data.overlap,
            "flux": data.flux,
            "startImageNumber": data.start_image_number,
            "resolution": data.resolution,
            "wavelength": data.wavelength,
            "xBeam": data.xbeam,
            "yBeam": data.ybeam,
            "synchrotronMode": data.synchrotron_mode,
            "undulatorGap1": data.undulator_gap1,
            "startTime": data.start_time,
            "endTime": data.end_time,
            "runStatus": data.run_status,
            "sampleId": data.sample_id,
            "detectorMode": data.detector_mode,
        }
    )


def _position_info_to_json(data: DataCollectionPositionInfo) -> dict:
    return _none_to_absent(
        {
            "posX": data.pos_x,
            "posY": data.pos_y,
            "posZ": data.pos_z,
        }
    )


def _grid_info_to_json(data: DataCollectionGridInfo) -> dict:
    return _none_to_absent(
        {
            "snapshotOffsetXPixel": data.snapshot_offset_x_pixel,
            "snapshotOffsetYPixel": data.snapshot_offset_y_pixel,
            "dx": data.dx_in_mm,
            "dy": data.dy_in_mm,
            "stepsX": data.steps_x,
            "stepsY": data.steps_y,
            "orientation": data.orientation.value,
            "micronsPerPixelX": data.microns_per_pixel_x,
            "micronsPerPixelY": data.microns_per_pixel_y,
            "snaked": data.snaked,
        }
    )



import bluesky.plan_stubs as bps
from dodal.devices.zebra.zebra import (
    ArmDemand,
    Zebra,
)
from dodal.devices.zebra.zebra_controlled_shutter import (
    ZebraShutter,
)

from mx_bluesky.common.device_setup_plans.setup_zebra_and_shutter import (
    configure_zebra_and_shutter_for_auto_shutter,
)
from mx_bluesky.common.parameters.constants import ZEBRA_STATUS_TIMEOUT


def arm_zebra(zebra: Zebra):
    yield from bps.abs_set(zebra.pc.arm, ArmDemand.ARM, wait=True)


def setup_zebra_for_panda_flyscan(
    zebra: Zebra,
    zebra_shutter: ZebraShutter,
    group="setup_zebra_for_panda_flyscan",
    wait=True,
):
    # Forwards eiger trigger signal from panda
    yield from bps.abs_set(
        zebra.output.out_pvs[zebra.mapping.outputs.TTL_DETECTOR],
        zebra.mapping.sources.IN1_TTL,
        group=group,
    )

    # Set shutter to automatic and to trigger via motion controller GPIO signal (IN4_TTL)
    yield from configure_zebra_and_shutter_for_auto_shutter(
        zebra, zebra_shutter, zebra.mapping.sources.IN4_TTL, group=group
    )

    yield from bps.abs_set(
        zebra.output.out_pvs[zebra.mapping.outputs.TTL_XSPRESS3],
        zebra.mapping.sources.DISCONNECT,
        group=group,
    )

    yield from bps.abs_set(
        zebra.output.out_pvs[zebra.mapping.outputs.TTL_PANDA],
        zebra.mapping.sources.IN3_TTL,
        group=group,
    )  # Tells panda that motion is beginning/changing direction

    if wait:
        yield from bps.wait(group, timeout=ZEBRA_STATUS_TIMEOUT)
